# -*- coding: utf-8 -*-
# Copyright (c) 2023, Johannes Heinecke.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .currency import parse_currency_parts
from .lang_EU import Num2Word_EU

# Welsh numerals differs to many other languages since the counted
# object does not follow the numeral but is inserted between
# e.g. "23 hours" is
#   tri awr ar hugain
#   3   hour on twenty
# in addition to that some numeral trigger a mutation on the following word
# either another numeral or the counted object
# (https://en.wikipedia.org/wiki/Consonant_mutation#Welsh)
# e.g. "23 dogs" (aspirated mutation, c -> ch)
#  tri chi ar hugain
#  3   dog on twenty
# but "22 dogs" (soft mutation, c -> g)
#  dau gi  ar hugain
#  2   dog on twenty
# and "24 dogs" (no mutation)
#  pedwar ci ar hugain
#  4      dog on twenty
# (BTW, the counted word is always in singular when following a numeral)
# numerals are mutated as well
# e.g. "300"
#    tri chant
#    3   hundred
# "200"
#    dau gant
#    2   hundred
# "500"
#    pump cant
#    5    hundreds
# the numerals for 2, 3 and 4 are different in function of gender (MASC, FEM)
#    2 cats
#    dwy gath

#    2 dogs
#    dau gi

#    2000
#    dwy fil

#    3000
#    tair mil

# to add the counted object in the correct position use
#   num2words(17, lang="cy", counted="ci", gender="masc")
#   num2words(17, lang="cy", counted="cath", gender="fem")
# if the number is > 99, use plural form of counted object
#   num2words(117, lang="cy", counted="cathod", gender="fem")


# Globals
# -------

OBJ = "__OBJECT__"

CARDINAL_WORDS = {
    # masc, fem, triggers mutation
    0: [("dim", None), (OBJ, None)],
    1: [("un", None), (OBJ, None)],
    2: [("dau", "SM"), (OBJ, None)],
    3: [("tri", "AM"), (OBJ, None)],
    4: [("pedwar", None), (OBJ, None)],
    5: [("pump", None), (OBJ, None)],
    6: [("chwech", "AM"), (OBJ, None)],
    7: [("saith", None), (OBJ, None)],
    8: [("wyth", None), (OBJ, None)],
    9: [("naw", None), (OBJ, None)],
    10: [("deg", None), (OBJ, None)],
    11: [("un", None), (OBJ, None), ("ar ddeg", None)],
    12: [("deuddeg", None), (OBJ, None)],
    13: [("tri", "AM"), (OBJ, None), ("ar ddeg", None)],
    14: [("pedwar", None), (OBJ, None), ("ar ddeg", None)],
    15: [("pymtheg", None), (OBJ, None)],
    16: [("un", None), (OBJ, None), ("ar bymtheg", None)],
    17: [("dau", "SM"), (OBJ, None), ("ar bymtheg", None)],
    18: [("deunaw", None), (OBJ, None)],
    19: [("pedwar", None), ("ar bymtheg", None)],
}

CARDINAL_WORDS_FEM = {
    # masc, fem, triggers mutation
    0: [("dim", None), (OBJ, None)],
    1: [("un", None), (OBJ, None)],
    2: [("dwy", "SM"), (OBJ, None)],
    3: [("tair", None), (OBJ, None)],
    4: [("pedair", None), (OBJ, None)],
    5: [("pump", None), (OBJ, None)],
    6: [("chwech", "AM"), (OBJ, None)],
    7: [("saith", None), (OBJ, None)],
    8: [("wyth", None), (OBJ, None)],
    9: [("naw", None), (OBJ, None)],
    10: [("deg", None), (OBJ, None)],
    11: [("un", None), (OBJ, None), ("ar ddeg", None)],
    12: [("deuddeg", None), (OBJ, None)],
    13: [("tair", None), (OBJ, None), ("ar ddeg", None)],
    14: [("pedair", None), (OBJ, None), ("ar ddeg", None)],
    15: [("pymtheg", None), (OBJ, None)],
    16: [("un", None), (OBJ, None), ("ar bymtheg", None)],
    17: [("dwy", "SM"), (OBJ, None), ("ar bymtheg", None)],
    18: [("deunaw", None), (OBJ, None)],
    19: [("pedair", None), ("ar bymtheg", None)],
}

MILLION_WORDS = {
    3: ("mil", None),
    6: ("miliwn", None),
    9: ("biliwn", None),
    12: ("triliwn", None),
    15: ("cwadriliwn", None),
    18: ("cwintiliwn", None),
    21: ("secsttiliwn", None),
    24: ("septiliwn", None),
    27: ("octiliwn", None),
    30: ("noniliwn", None),
    33: ("dengiliwn", None),
}

ORDINAL_WORDS = {
    0: [("dimfed", None), (OBJ, None)],
    1: [(OBJ, None), ("cyntaf", None)],
    2: [("ail", "SM"), (OBJ, None)],
    3: [("trydydd", None), (OBJ, None)],
    4: [("pedwerydd", None), (OBJ, None)],
    5: [("pumed", None), (OBJ, None)],
    6: [("chweched", None), (OBJ, None)],
    7: [("saithfed", None), (OBJ, None)],
    8: [("wythfed", None), (OBJ, None)],
    9: [("nawfed", None), (OBJ, None)],
    10: [("degfed", None), (OBJ, None)],
    11: [("unfed", "SM"), (OBJ, None), ("ar ddeg", None)],
    12: [("deuddegfed", None), (OBJ, None)],
    13: [("trydydd", None), (OBJ, None), ("ar ddeg", None)],
    14: [("pedwerydd", None), (OBJ, None), ("ar ddeg", None)],
    15: [("pymthegfed", None), (OBJ, None)],
    16: [("unfed", None), (OBJ, None), ("ar bymtheg", None)],
    17: [("ail", "SM"), (OBJ, None), ("ar bymtheg", None)],
    18: [("deunawfed", None), (OBJ, None)],
    19: [("pedwerydd", None), (OBJ, None), ("ar bymtheg", None)],
}
ORDINAL_WORDS_FEM = {
    0: [("dimfed", None), (OBJ, None)],
    1: [(OBJ, None), ("gyntaf", None)],
    2: [("ail", "SM"), (OBJ, None)],
    3: [("trydedd", "SM"), (OBJ, None)],
    4: [("pedwaredd", "SM"), (OBJ, None)],
    5: [("pumed", None), (OBJ, None)],
    6: [("chweched", None), (OBJ, None)],
    7: [("saithfed", None), (OBJ, None)],
    8: [("wythfed", None), (OBJ, None)],
    9: [("nawfed", None), (OBJ, None)],
    10: [("degfed", None), (OBJ, None)],
    11: [("unfed", "SM"), (OBJ, None), ("ar ddeg", None)],
    12: [("deuddegfed", None), (OBJ, None)],
    13: [("trydedd", "SM"), (OBJ, None), ("ar ddeg", None)],
    14: [("pedwaredd", "SM"), (OBJ, None), ("ar ddeg", None)],
    15: [("pymthegfed", None), (OBJ, None)],
    16: [("unfed", None), (OBJ, None), ("ar bymtheg", None)],
    17: [("ail", "SM"), (OBJ, None), ("ar bymtheg", None)],
    18: [("deunawfed", None), (OBJ, None)],
    19: [("pedwaredd", None), (OBJ, None), ("ar bymtheg", None)],
}

# The script can extrapolate the missing numbers from the base forms.
STR_TENS = {
    1: [("ugain", None), (OBJ, None)],
    2: [("deugain", None), (OBJ, None)],
    3: [("trigain", None), (OBJ, None)],
    4: [("pedwar ugain", None), (OBJ, None)],
}

ORD_STR_TENS = {
    1: [("ugainfed", None), (OBJ, None)],
    2: [("deugainfed", None), (OBJ, None)],
    3: [("trigainfed", None), (OBJ, None)],
    4: [("pedwar ugainfed", None), (OBJ, None)],
}

STR_TENS_INFORMAL = {
    1: ("undeg", None),
    2: ("dauddeg", None),
    3: ("trideg", None),
    4: ("pedwardeg", None),
    5: ("pumdeg", None),
    6: ("chwedeg", None),
    7: ("saithdeg", None),
    8: ("wythdeg", None),
    9: ("nawdeg", None),
}


GENERIC_DOLLARS = ("dolar", "dolarau")
GENERIC_CENTS = ("ceiniog", "ceiniogau")

CURRENCIES_FEM = ["GBP"]


class Num2Word_CY(Num2Word_EU):
    CURRENCY_FORMS = {
        # currency code: (sg, pl), (sg, pl)
        # in Welsh a noun after a numeral is ALWAYS in the singular
        "EUR": (("euro", "euros"), GENERIC_CENTS),
        "USD": (GENERIC_DOLLARS, GENERIC_CENTS),
        "GBP": (("punt", "punnoedd"), ("ceiniog", "ceiniogau")),
        "CNY": (("yuan", "yuans"), ("ffen", "ffens")),
    }

    MINUS_PREFIX_WORD = "meinws "
    FLOAT_INFIX_WORD = " pwynt "

#    def setup(self):
#        Num2Word_EU.setup(self)

    def __init__(self):
        pass

    def float_to_words(self, float_number):
        # if ordinal:
        #     prefix = self.to_ordinal(int(float_number))
        # else:
        prefix = self.to_cardinal(int(float_number))
        float_part = str(float_number).split(".")[1]
        postfix = " ".join(
            # Drops the trailing zero and comma
            [self.to_cardinal(int(c)) for c in float_part]
        )
        return prefix + Num2Word_CY.FLOAT_INFIX_WORD + postfix

    def hundred_group(
        self, number, informal=False, gender="masc", ordinal=False
    ):
        hundreds = number // 100
        until100 = number % 100  # 0 - 99
        # list group of number words and mutation info (for the following word)
        result = (
            []
        )
        if gender == "fem":
            CW = CARDINAL_WORDS_FEM
        else:
            if ordinal:
                CW = ORDINAL_WORDS
            else:
                CW = CARDINAL_WORDS

        if hundreds > 0:
            if hundreds > 1:
                result.extend((CARDINAL_WORDS[hundreds]))
            result.extend([("cant", None), (OBJ, None)])
            if until100:
                if until100 in [
                    1,
                    8,
                    11,
                    16,
                    20,
                    21,
                    31,
                    36,
                    41,
                    48,
                    61,
                    68,
                    71,
                    81,
                    88,
                    91,
                ]:
                    result.append(("ac", None))
                else:
                    result.append(("a", "AM"))
        if until100:
            # if informal:
            #    pass
            if not ordinal and until100 >= 50 and until100 <= 59:
                units = number % 10
                if hundreds > 0:
                    if units == 0:
                        result.append(("hanner", None))
                    elif units == 1:
                        result.extend([("hanner ac un", None), (OBJ, None)])
                    else:
                        result.append(("hanner a", "AM"))
                        result.extend(CW[units])
                else:
                    if units == 0:
                        result.extend([("hanner cant", None), (OBJ, None)])
                    elif units == 1:
                        result.extend(
                            [("hanner cant ac un", None), (OBJ, None)]
                        )
                    else:
                        result.append(("hanner cant a", "AM"))
                        result.extend(CW[units])
            else:
                if (number < 20 and number > 0) or (
                    number == 0 and hundreds == 0
                ):
                    if gender == "fem":
                        result.extend(CARDINAL_WORDS_FEM[int(number)])
                    else:
                        result.extend(CARDINAL_WORDS[int(number)])

                else:
                    tens = until100 // 20
                    units = number % 20
                    if ordinal and units == 0:
                        degau = ORD_STR_TENS.get(tens)
                    else:
                        degau = STR_TENS.get(tens)

                    if units != 0:
                        if tens > 1:
                            result.extend(CW[units])
                            if degau:
                                result.append(("a", "AM"))
                                result.extend(degau)
                        else:
                            result.extend(CW[units])
                            if degau:
                                result.append(("ar", "SM"))
                                result.extend(degau)
                    elif degau:
                        result.extend(degau)
        return result

    def to_ordinal(self, number, informal=False, gender="masc"):
        if number < 20:
            return makestring(ORDINAL_WORDS[number])
        if number == 100:
            return "canfed"
        elif number > 100:
            raise NotImplementedError("The given number is too large.")

        return self.to_cardinal(
            number, informal=False, gender=gender, ordinal=True
        )

    def to_cardinal(
        self,
        number,
        informal=False,
        gender="masc",
        ordinal=False,
        counted=None,
        raw=False,
    ):
        negative = False
        if number < 0:
            negative = True
            number = -1 * number
        if number == 0:
            if raw:
                return CARDINAL_WORDS[0]
            else:
                return makestring(CARDINAL_WORDS[0])
        elif not number < 999 * 10**33:
            raise NotImplementedError("The given number is too large.")

        elif isinstance(number, float):
            return self.float_to_words(number)

        # split in groups of 10**3
        # groups of three digits starting from right (units (1 - 999),
        # thousands, millions, ...)
        groups = (
            []
        )
        lowestgroup = (
            None  # find the lowest group of 3 digits > 0 for the ordinals
        )
        for pot in [3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 36]:
            gr = (number % 10**pot) // 10 ** (pot - 3)
            groups.append((gr, pot))
            if gr and not lowestgroup:
                lowestgroup = gr
        # print("groups", groups)

        result = []
        if negative:
            result.append(("meinws", None))

        for gr, pot in reversed(groups):
            if gr:
                # print("AAAA", gr, pot, gender)
                if pot == 6:
                    g = "fem"  # mil (1000) is feminine
                elif pot == 3:
                    g = gender  # units depend on the following noun
                else:
                    g = "masc"  # millions etc are masculine
                    # "mil" is feminine
                if gr > 1 or pot == 3:
                    words = self.hundred_group(
                        gr,
                        informal=informal,
                        gender=g,
                        ordinal=ordinal and (lowestgroup == gr),
                    )
                    result += words
                    # print(">>>> ", words)
                if pot > 3:
                    result.append(MILLION_WORDS[pot - 3])
        if raw:
            # need to be able trigger correct mutation on currencies
            return result
        else:
            if number < 100:
                return makestring(result, counted=counted)
            else:
                if counted:
                    result.extend([("o", "SM"), (counted, None)])
                return makestring(result)

    def to_currency(
        self, val, currency="EUR", cents=True, separator=",", adjective=False
    ):
        """
        Args:
            val: Numeric value
            currency (str): Currency code
            cents (bool): Verbose cents
            separator (str): Cent separator
            adjective (bool): Prefix currency name with adjective
        Returns:
            str: Formatted string

        """
        left, right, is_negative = parse_currency_parts(val)
        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]

        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"'
                % (currency, self.__class__.__name__)
            )

        # if adjective and currency in self.CURRENCY_ADJECTIVES:
        #     cr1 = prefix_currency(self.CURRENCY_ADJECTIVES[currency], cr1)

        minus_str = "%s " % self.negword.strip() if is_negative else ""
        money_str = self._money_verbose(left, currency)
        cents_str = (
            self._cents_verbose(right, currency)
            if cents
            else self._cents_terse(right, currency)
        )

        if right == 0:
            # no pence
            return "%s%s" % (
                minus_str,
                money_str,
                # self.pluralize(right, cr2)
            )
        elif left == 0:
            # no pounds
            return "%s%s" % (
                minus_str,
                cents_str,
                # self.pluralize(right, cr2)
            )

        return "%s%s%s %s" % (
            minus_str,
            money_str,
            # self.pluralize(left, cr1),
            separator,
            cents_str,
            # self.pluralize(right, cr2)
        )

    def _money_verbose(self, number, currency):
        # used in super().to_currency(), we need to add gender
        # here for feminine currencies
        # if currency in CURRENCIES_FEM:  # always true in this context
        if number > 100:
            m = self.to_cardinal(number, gender="fem", raw=True)
            # if currency in self.CURRENCY_FORMS:
            c = self.CURRENCY_FORMS[currency][0][1]
            m.append(("o", "SM"))
            m.append((c, None))
            # else:
            #    c = currency
            #    m.append((c, None))
            return makestring(m)
        else:
            # if number > 1:
            m = self.to_cardinal(number, gender="fem", raw=True)
            # elif number == 0:
            #    m = self.to_cardinal(number, gender="fem", raw=True)
            # else:
            #    m = [(OBJ, None)]
            # if currency in self.CURRENCY_FORMS:
            c = self.CURRENCY_FORMS[currency][0][0]
            # else:
            #    c = currency
            # print("eeeeeeeee", m)
            # m.append((c, None))
            # print("fffffffff", m)
            return makestring(m, counted=c)
        # else:
        #    return self.to_cardinal(number, raw=True)

    def _cents_verbose(self, number, currency):
        if number == 0:
            return ""
        # elif number > 100:
        #     m = self.to_cardinal(number, raw=True)
        #     # if currency in self.CURRENCY_FORMS:
        #     c = self.CURRENCY_FORMS[currency][0][1]
        #     m.append(("o", "SM"))
        #     m.append((c, None))
        #     # else:
        #     #    c = currency
        #     #    m.append((c, None))
        #     return makestring(m)
        else:
            if number > 1:
                m = self.to_cardinal(number, raw=True)
            else:
                m = [(OBJ, None)]
            # if currency in self.CURRENCY_FORMS:
            c = self.CURRENCY_FORMS[currency][1][0]
            # else:
            #    c = currency
            return makestring(m, counted=c)


def makestring(result, counted=None):
    # concatenate numberwords with correct mutation
    out = []
    lastmut = None
    for w, mut in result:
        if w == OBJ:
            if not counted:
                continue
            else:
                w = counted
                counted = None  # only first position
        if lastmut:
            out.append(mutate(w, lastmut))
        else:
            out.append(w)
        lastmut = mut
    return " ".join(out)


def mutate(word, mutation):
    # print("uuu", word, mutation)
    if mutation == "SM":
        return softmutation(word)
    elif mutation == "AM":
        return aspiratedmutation(word)
    # return word  # does not occur


def softmutation(word):
    # print("SM<<<<%s>" % word)
    if word[0] == "p" and word[1] != "h":
        return "b" + word[1:]
    elif word[0] == "t" and word[1] != "h":
        return "d" + word[1:]
    elif word[0] == "c" and word[1] != "h":
        return "g" + word[1:]
    elif word[0] == "b" or word[0] == "m":
        return "f" + word[1:]
    elif word[0] == "d" and word[1] != "d":
        return "d" + word
    elif word.startswith("ll"):
        return word[1:]
    elif word.startswith("rh"):
        return "r" + word[2:]
    elif word == "ugain":
        return "hugain"
    else:
        return word


def aspiratedmutation(word):
    if word[0] == "p" and word[1] != "h":
        return "ph" + word[1:]
    elif word[0] == "t" and word[1] != "h":
        return "th" + word[1:]
    elif word[0] == "c" and word[1] != "h":
        return "ch" + word[1:]
    else:
        return word



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base


class Num2Word_AZ(Num2Word_Base):
    DIGITS = {
        0: u"sıfır",
        1: u"bir",
        2: u"iki",
        3: u"üç",
        4: u"dörd",
        5: u"beş",
        6: u"altı",
        7: u"yeddi",
        8: u"səkkiz",
        9: u"doqquz",
    }

    DECIMALS = {
        1: u"on",
        2: u"iyirmi",
        3: u"otuz",
        4: u"qırx",
        5: u"əlli",
        6: u"altmış",
        7: u"yetmiş",
        8: u"səksən",
        9: u"doxsan",
    }

    POWERS_OF_TEN = {
        2: u"yüz",
        3: u"min",
        6: u"milyon",
        9: u"milyard",
        12: u"trilyon",
        15: u"katrilyon",
        18: u"kentilyon",
        21: u"sekstilyon",
        24: u"septilyon",
        27: u"oktilyon",
        30: u"nonilyon",
        33: u"desilyon",
        36: u"undesilyon",
        39: u"dodesilyon",
        42: u"tredesilyon",
        45: u"katordesilyon",
        48: u"kendesilyon",
        51: u"seksdesilyon",
        54: u"septendesilyon",
        57: u"oktodesilyon",
        60: u"novemdesilyon",
        63: u"vigintilyon",
    }

    VOWELS = u"aıoueəiöü"
    VOWEL_TO_CARDINAL_SUFFIX_MAP = {
        **dict.fromkeys(["a", "ı"], "ıncı"),
        **dict.fromkeys(["e", "ə", "i"], "inci"),
        **dict.fromkeys(["o", "u"], "uncu"),
        **dict.fromkeys(["ö", "ü"], "üncü"),
    }

    VOWEL_TO_CARDINAL_NUM_SUFFIX_MAP = {
        **dict.fromkeys(["a", "ı"], "cı"),
        **dict.fromkeys(["e", "ə", "i"], "ci"),
        **dict.fromkeys(["o", "u"], "cu"),
        **dict.fromkeys(["ö", "ü"], "cü"),
    }

    CURRENCY_INTEGRAL = ('manat', 'manat')
    CURRENCY_FRACTION = ('qəpik', 'qəpik')
    CURRENCY_FORMS = {'AZN': (CURRENCY_INTEGRAL, CURRENCY_FRACTION)}

    def setup(self):
        super().setup()

        self.negword = u"mənfi"
        self.pointword = u"nöqtə"

    def to_cardinal(self, value):
        value_str = str(value)
        parts = value_str.split(".")
        integral_part = parts[0]
        fraction_part = parts[1] if len(parts) > 1 else ""

        if integral_part.startswith("-"):
            integral_part = integral_part[1:]  # ignore minus sign here

        integral_part_in_words = self.int_to_word(integral_part)
        fraction_part_in_words = "" if not fraction_part else self.int_to_word(
            fraction_part, leading_zeros=True)

        value_in_words = integral_part_in_words
        if fraction_part:
            value_in_words = " ".join(
                [
                    integral_part_in_words,
                    self.pointword,
                    fraction_part_in_words
                ]
            )

        if value < 0:
            value_in_words = " ".join([self.negword, value_in_words])

        return value_in_words

    def to_ordinal(self, value):
        assert int(value) == value

        cardinal = self.to_cardinal(value)
        last_vowel = self._last_vowel(cardinal)
        assert last_vowel is not None
        suffix = self.VOWEL_TO_CARDINAL_SUFFIX_MAP[last_vowel]

        if cardinal.endswith(tuple(self.VOWELS)):
            cardinal = cardinal[:-1]

        ordinal = "".join([cardinal, suffix])

        return ordinal

    def to_ordinal_num(self, value):
        assert int(value) == value

        cardinal = self.to_cardinal(value)
        last_vowel = self._last_vowel(cardinal)
        assert last_vowel is not None
        suffix = self.VOWEL_TO_CARDINAL_NUM_SUFFIX_MAP[last_vowel]
        ordinal_num = "-".join([str(value), suffix])

        return ordinal_num

    def to_year(self, value):
        assert int(value) == value

        year = self.to_cardinal(abs(value))
        if value < 0:
            year = " ".join(["e.ə.", year])

        return year

    def pluralize(self, n, forms):
        form = 0 if n == 1 else 1
        return forms[form]

    def int_to_word(self, num_str, leading_zeros=False):
        words = []
        reversed_str = list(reversed(num_str))

        for index, digit in enumerate(reversed_str):
            digit_int = int(digit)
            # calculate remainder after dividing index by 3
            # because number is parsed by three digit chunks
            remainder_to_3 = index % 3
            if remainder_to_3 == 0:
                if index > 0:
                    if set(reversed_str[index:index+3]) != {'0'}:
                        words.insert(0, self.POWERS_OF_TEN[index])
                if digit_int > 0:
                    # we say "min" not "bir min"
                    if not (digit_int == 1 and index == 3):
                        words.insert(0, self.DIGITS[digit_int])
            elif remainder_to_3 == 1:
                if digit_int != 0:
                    words.insert(0, self.DECIMALS[digit_int])
            else:  # remainder is 2
                if digit_int > 0:
                    words.insert(0, self.POWERS_OF_TEN[2])  # "yüz"
                if digit_int > 1:
                    words.insert(0, self.DIGITS[digit_int])

        if num_str == '0':
            words.append(self.DIGITS[0])

        if leading_zeros:
            zeros_count = len(num_str) - len(str(int(num_str)))
            words[:0] = zeros_count * [self.DIGITS[0]]

        return " ".join(words)

    def _last_vowel(self, value):
        for char in value[::-1]:
            if char in self.VOWELS:
                return char



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

ZERO = 'нөл'

ONES = {
    1: 'бір',
    2: 'екі',
    3: 'үш',
    4: 'төрт',
    5: 'бес',
    6: 'алты',
    7: 'жеті',
    8: 'сегіз',
    9: 'тоғыз',
}

TEN = 'он'

TWENTIES = {
    2: 'жиырма',
    3: 'отыз',
    4: 'қырық',
    5: 'елу',
    6: 'алпыс',
    7: 'жетпіс',
    8: 'сексен',
    9: 'тоқсан',
}

HUNDRED = 'жүз'

THOUSANDS = {
    1: 'мың',
    2: 'миллион',
    3: 'миллиард',
    4: 'триллион',
    5: 'квадриллион',
    6: 'квинтиллион',
    7: 'секстиллион',
    8: 'септиллион',
    9: 'октиллион',
    10: 'нониллион',
}


class Num2Word_KZ(Num2Word_Base):
    CURRENCY_FORMS = {
        'USD': ('доллар', 'цент'),
        'KZT': ('теңге', 'тиын'),
    }

    def setup(self):
        self.negword = "минус"
        self.pointword = "бүтін"

    def to_cardinal(self, number):
        n = str(number).replace(',', '.')
        if '.' in n:
            left, right = n.split('.')
            leading_zero_count = len(right) - len(right.lstrip('0'))
            return u'%s %s %s' % (
                self._int2word(int(left)),
                self.pointword,
                (ZERO + ' ') * leading_zero_count + self._int2word(int(right))
            )
        else:
            return self._int2word(int(n))

    def pluralize(self, n, form):
        return form

    def _cents_verbose(self, number, currency):
        return self._int2word(number, currency == 'KZT')

    def _int2word(self, n, feminine=False):
        if n < 0:
            return ' '.join([self.negword, self._int2word(abs(n))])

        if n == 0:
            return ZERO

        words = []
        chunks = list(splitbyx(str(n), 3))
        i = len(chunks)
        for x in chunks:
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            if n3 > 0:
                if n3 > 1:
                    words.append(ONES[n3])
                words.append(HUNDRED)

            if n2 == 1:
                words.append(TEN)
            elif n2 > 1:
                words.append(TWENTIES[n2])

            if n1 > 0:
                words.append(ONES[n1])

            if i > 0:
                words.append(THOUSANDS[i])

        return ' '.join(words)

    def to_ordinal(self, number):
        # TODO: Implement to_ordinal
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU


class Num2Word_AM(lang_EU.Num2Word_EU):
    CURRENCY_FORMS = {'ETB': (('ብር', 'ብር'), ('ሳንቲም', 'ሳንቲም'))}

    GIGA_SUFFIX = 'ቢሊዮን'
    MEGA_SUFFIX = 'ሚሊዮን'

    def set_high_numwords(self, high):
        cap = 3 * (len(high) + 1)

        for word, n in zip(high, range(cap, 5, -3)):
            if n == 9:
                self.cards[10 ** n] = word + self.GIGA_SUFFIX
            else:
                self.cards[10 ** n] = word + self.MEGA_SUFFIX

    def setup(self):
        super(Num2Word_AM, self).setup()

        self.negword = 'አሉታዊ '
        self.pointword = 'ነጥብ'
        self.exclude_title = ['እና', 'ነጥብ', 'አሉታዊ']

        self.mid_numwords = [(1000, 'ሺህ'), (100, 'መቶ'), (90, 'ዘጠና'),
                             (80, 'ሰማኒያ'), (70, 'ሰባ'), (60, 'ስድሳ'),
                             (50, 'አምሳ'), (40, 'አርባ'), (30, 'ሠላሳ')]
        self.low_numwords = ['ሃያ', 'አሥራ ዘጠኝ', 'አሥራ ስምንት', 'አሥራ ሰባት',
                             'አስራ ስድስት', 'አሥራ አምስት', 'አሥራ አራት', 'አሥራ ሦስት',
                             'አሥራ ሁለት', 'አሥራ አንድ', 'አሥር', 'ዘጠኝ', 'ስምንት',
                             'ሰባት', 'ስድስት', 'አምስት', 'አራት', 'ሦስት', 'ሁለት',
                             'አንድ', 'ዜሮ']
        self.ords = {'አንድ': 'አንደኛ',
                     'ሁለት': 'ሁለተኛ',
                     'ሦስት': 'ሦስተኛ',
                     'አራት': 'አራተኛ',
                     'አምስት': 'አምስተኛ',
                     'ስድስት': 'ስድስተኛ',
                     'ሰባት': 'ሰባተኛ',
                     'ስምንት': 'ስምንተኛ',
                     'ዘጠኝ': 'ዘጠነኛ',
                     'አሥር': 'አሥረኛ',
                     'አሥራ አንድ': 'አሥራ አንደኛ',
                     'አሥራ ሁለት': 'አሥራ ሁለተኛ',
                     'አሥራ ሦስት': 'አሥራ ሦስተኛ',
                     'አሥራ አራት': 'አሥራ አራተኛ',
                     'አሥራ አምስት': 'አሥራ አምስተኛ',
                     'አሥራ ስድስት': 'አሥራ ስድስተኛ',
                     'አሥራ ሰባት': 'አሥራ ሰባተኛ',
                     'አሥራ ስምንት': 'አሥራ ስምንተኛ',
                     'አሥራ ዘጠኝ': 'አሥራ ዘጠነኛ'}

    def to_cardinal(self, value):
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            return self.to_cardinal_float(value)

        out = ''
        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        if value == 100:
            return self.title(out + 'መቶ')
        else:
            val = self.splitnum(value)
            words, num = self.clean(val)
            return self.title(out + words)

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return rtext, rnum
        elif 100 > lnum > rnum:
            return '%s %s' % (ltext, rtext), lnum + rnum
        elif lnum >= 100 > rnum:
            return '%s %s' % (ltext, rtext), lnum + rnum
        elif rnum > lnum:
            return '%s %s' % (ltext, rtext), lnum * rnum

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outwords = self.to_cardinal(value).split(' ')
        lastwords = outwords[-1].split('-')
        lastword = lastwords[-1].lower()
        try:
            lastword = self.ords[lastword]
        except KeyError:
            lastword += 'ኛ'
        lastwords[-1] = self.title(lastword)
        outwords[-1] = ' '.join(lastwords)
        return ' '.join(outwords)

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return '%s%s' % (value, self.to_ordinal(value)[-1:])

    def to_currency(self, val, currency='ብር', cents=True, separator=' ከ',
                    adjective=True):
        result = super(Num2Word_AM, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)
        return result

    def to_year(self, val, longval=True):
        if not (val // 100) % 10:
            return self.to_cardinal(val)
        return self.to_splitnum(val, hightxt='መቶ', longval=longval)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

import itertools

from .base import Num2Word_Base
from .utils import get_digits, splitbyx

ZERO = ('zero',)

ONES = {
    1: ('jeden',),
    2: ('dwa',),
    3: ('trzy',),
    4: ('cztery',),
    5: ('pięć',),
    6: ('sześć',),
    7: ('siedem',),
    8: ('osiem',),
    9: ('dziewięć',),
}

ONES_ORDINALS = {
    1: ('pierwszy', "pierwszo"),
    2: ('drugi', "dwu"),
    3: ('trzeci', "trzy"),
    4: ('czwarty', "cztero"),
    5: ('piąty', "pięcio"),
    6: ('szósty', "sześcio"),
    7: ('siódmy', "siedmio"),
    8: ('ósmy', "ośmio"),
    9: ('dziewiąty', "dziewięcio"),
    10: ('dziesiąty', "dziesięcio"),
    11: ('jedenasty', "jedenasto"),
    12: ('dwunasty', "dwunasto"),
    13: ('trzynasty', "trzynasto"),
    14: ('czternasty', "czternasto"),
    15: ('piętnasty', "piętnasto"),
    16: ('szesnasty', "szesnasto"),
    17: ('siedemnasty', "siedemnasto"),
    18: ('osiemnasty', "osiemnasto"),
    19: ('dziewiętnasty', "dziewiętnasto"),
}

TENS = {
    0: ('dziesięć',),
    1: ('jedenaście',),
    2: ('dwanaście',),
    3: ('trzynaście',),
    4: ('czternaście',),
    5: ('piętnaście',),
    6: ('szesnaście',),
    7: ('siedemnaście',),
    8: ('osiemnaście',),
    9: ('dziewiętnaście',),
}


TWENTIES = {
    2: ('dwadzieścia',),
    3: ('trzydzieści',),
    4: ('czterdzieści',),
    5: ('pięćdziesiąt',),
    6: ('sześćdziesiąt',),
    7: ('siedemdziesiąt',),
    8: ('osiemdziesiąt',),
    9: ('dziewięćdziesiąt',),
}

TWENTIES_ORDINALS = {
    2: ('dwudziesty', "dwudziesto"),
    3: ('trzydziesty', "trzydiesto"),
    4: ('czterdziesty', "czterdziesto"),
    5: ('pięćdziesiąty', "pięćdziesięcio"),
    6: ('sześćdziesiąty', "sześćdziesięcio"),
    7: ('siedemdziesiąty', "siedemdziesięcio"),
    8: ('osiemdziesiąty', "osiemdziesięcio"),
    9: ('dziewięćdziesiąty', "dziewięćdziesięcio"),
}

HUNDREDS = {
    1: ('sto',),
    2: ('dwieście',),
    3: ('trzysta',),
    4: ('czterysta',),
    5: ('pięćset',),
    6: ('sześćset',),
    7: ('siedemset',),
    8: ('osiemset',),
    9: ('dziewięćset',),
}

HUNDREDS_ORDINALS = {
    1: ('setny', "stu"),
    2: ('dwusetny', "dwustu"),
    3: ('trzysetny', "trzystu"),
    4: ('czterysetny', "czterystu"),
    5: ('pięćsetny', "pięcset"),
    6: ('sześćsetny', "sześćset"),
    7: ('siedemsetny', "siedemset"),
    8: ('osiemsetny', "ośiemset"),
    9: ('dziewięćsetny', "dziewięćset"),
}

THOUSANDS = {
    1: ('tysiąc', 'tysiące', 'tysięcy'),  # 10^3
}

prefixes_ordinal = {
    1: "tysięczny",
    2: "milionowy",
    3: "milairdowy"
}

prefixes = (   # 10^(6*x)
    "mi",      # 10^6
    "bi",      # 10^12
    "try",     # 10^18
    "kwadry",  # 10^24
    "kwinty",  # 10^30
    "seksty",  # 10^36
    "septy",   # 10^42
    "okty",    # 10^48
    "nony",    # 10^54
    "decy"     # 10^60
)
suffixes = ("lion", "liard")  # 10^x or 10^(x+3)

for idx, (p, s) in enumerate(itertools.product(prefixes, suffixes)):
    name = p + s
    THOUSANDS[idx+2] = (name, name + 'y', name + 'ów')


class Num2Word_PL(Num2Word_Base):
    CURRENCY_FORMS = {
        'PLN': (
            ('złoty', 'złote', 'złotych'), ('grosz', 'grosze', 'groszy')
        ),
        'EUR': (
            ('euro', 'euro', 'euro'), ('cent', 'centy', 'centów')
        ),
        'USD': (
            (
                'dolar amerykański',
                'dolary amerykańskie',
                'dolarów amerykańskich'
            ),
            (
                'cent',
                'centy',
                'centów'
            )
        ),
    }

    def setup(self):
        self.negword = "minus"
        self.pointword = "przecinek"

    def to_cardinal(self, number):
        n = str(number).replace(',', '.')
        if '.' in n:
            left, right = n.split('.')
            leading_zero_count = len(right) - len(right.lstrip('0'))
            decimal_part = ((ZERO[0] + ' ') * leading_zero_count +
                            self._int2word(int(right)))
            return u'%s %s %s' % (
                self._int2word(int(left)),
                self.pointword,
                decimal_part
            )
        else:
            return self._int2word(int(n))

    def pluralize(self, n, forms):
        if n == 1:
            form = 0
        elif 5 > n % 10 > 1 and (n % 100 < 10 or n % 100 > 20):
            form = 1
        else:
            form = 2
        return forms[form]

    def last_fragment_to_ordinal(self, last, words, level):
        n1, n2, n3 = get_digits(last)
        last_two = n2*10+n1
        if last_two == 0:
            words.append(HUNDREDS_ORDINALS[n3][level])
        elif level == 1 and last == 1:
            return
        elif last_two < 20:
            if n3 > 0:
                words.append(HUNDREDS[n3][level])
            words.append(ONES_ORDINALS[last_two][level])
        elif last_two % 10 == 0:
            if n3 > 0:
                words.append(HUNDREDS[n3][level])
            words.append(TWENTIES_ORDINALS[n2][level])
        else:
            if n3 > 0:
                words.append(HUNDREDS[n3][0])
            words.append(TWENTIES_ORDINALS[n2][0])
            words.append(ONES_ORDINALS[n1][0])

    def to_ordinal(self, number):
        if number % 1 != 0:
            raise NotImplementedError()
        words = []
        fragments = list(splitbyx(str(number), 3))
        level = 0
        last = fragments[-1]
        while last == 0:
            level = level+1
            fragments.pop()
            last = fragments[-1]
        if len(fragments) > 1:
            pre_part = self._int2word(number-(last*1000**level))
            words.append(pre_part)
        self.last_fragment_to_ordinal(last, words, 0 if level == 0 else 1)
        output = " ".join(words)
        if last == 1 and level > 0 and output != "":
            output = output + " "
        if level > 0:
            output = output + prefixes_ordinal[level]
        return output

    def _int2word(self, n):
        if n == 0:
            return ZERO[0]

        words = []
        chunks = list(splitbyx(str(n), 3))
        i = len(chunks)
        for x in chunks:
            i -= 1

            if x == 0:
                continue

            n1, n2, n3 = get_digits(x)

            if n3 > 0:
                words.append(HUNDREDS[n3][0])

            if n2 > 1:
                words.append(TWENTIES[n2][0])

            if n2 == 1:
                words.append(TENS[n1][0])
            elif n1 > 0 and not (i > 0 and x == 1):
                words.append(ONES[n1][0])

            if i > 0:
                words.append(self.pluralize(x, THOUSANDS[i]))

        return ' '.join(words)
